import java.util.*;
import java.io.*;
import localsolver.*;

public class FlexibleCumulative {
    // Number of tasks
    private int nbTasks;
    // Number of resources
    private int nbResources;
    // Maximum capacity of each resource
    private int[] capacity;
    // Duration of task i if task i is done by resource r
    private int[][] taskProcessingTimeData;
    // Resource weight of resource r required for task i
    private int[][] weightData;
    // Number of successors
    private int[] nbSuccessors;
    // Successors for each task i
    private int[][] successors;
    // Trivial upper bound for the start times of the tasks
    private int horizon = 0;

    // LocalSolver
    final LocalSolver localsolver;
    // Decision variables: set of tasks done by each resource
    private LSExpression[] resourcesTasks;
    // Decision variables: time range of each task
    private LSExpression[] tasks;
    // For each task, the selected resource
    private LSExpression[] taskResource;
    // Objective = minimize the makespan: end of the last task
    private LSExpression makespan;

    public FlexibleCumulative(LocalSolver localsolver, String fileName) throws IOException {
        this.localsolver = localsolver;
    }

    private void readInstance(String fileName) throws IOException {
        try (Scanner input = new Scanner(new File(fileName))) {
            nbTasks = input.nextInt();
            nbResources = input.nextInt();

            capacity = new int[nbResources];
            for (int r = 0; r < nbResources; ++r) {
                capacity[r] = input.nextInt();
            }

            taskProcessingTimeData = new int[nbTasks][nbResources];
            weightData = new int[nbResources][nbTasks];
            nbSuccessors = new int[nbTasks];
            successors = new int[nbTasks][];
            for (int i = 0; i < nbTasks; ++i) {
                for (int r = 0; r < nbResources; ++r) {
                    taskProcessingTimeData[i][r] = input.nextInt();
                    weightData[r][i] = input.nextInt();
                }
            }
            for (int i = 0; i < nbTasks; ++i) {
                nbSuccessors[i] = input.nextInt();
                successors[i] = new int[nbSuccessors[i]];
                for (int s = 0; s < nbSuccessors[i]; ++s) {
                    successors[i][s] = input.nextInt();
                }

                int maxDurationTask = taskProcessingTimeData[i][0];
                for (int r = 0; r < nbResources; ++r) {
                    if (taskProcessingTimeData[i][r] > taskProcessingTimeData[i][0]) {
                        maxDurationTask = taskProcessingTimeData[i][r];
                    }
                }
                horizon += maxDurationTask;
            }
        }
    }

    public void solve(int timeLimit) {
        // Declare the optimization model
        LSModel model = localsolver.getModel();

        resourcesTasks = new LSExpression[nbResources];

        for (int r = 0; r < nbResources; ++r) {
            resourcesTasks[r] = model.setVar(nbTasks);
        }
        // Create LocalSolver arrays to be able to access them with "at" operators
        LSExpression resources = model.array(resourcesTasks);
        LSExpression weight = model.array(weightData);
        LSExpression taskProcessingTime = model.array(taskProcessingTimeData);

        // Only compatible resources can be selected for a task
        for (int i = 0; i < nbTasks; ++i) {
            for (int r = 0; r < nbResources; ++r) {
                if (taskProcessingTimeData[i][r] == 0 && weightData[r][i] == 0) {
                    model.constraint(model.eq(model.contains(resourcesTasks[r], i), 0));
                }
            }
        }
        
        // All tasks are scheduled on the resources
        model.constraint(model.partition(resources));
        
        taskResource = new LSExpression[nbTasks];
        for (int i = 0; i < nbTasks; ++i) {
            // For each task, the selected resource
            taskResource[i] = model.find(resources, i);
        }

        tasks = new LSExpression[nbTasks];
        for (int i = 0; i < nbTasks; ++i) {
            // Interval decisions: time range of each task
            tasks[i] = model.intervalVar(0, horizon);

            // Task duration constraints
            LSExpression iExpr = model.createConstant(i);
            LSExpression duration = model.at(taskProcessingTime, iExpr, taskResource[i]);
            model.constraint(model.eq(model.length(tasks[i]), duration));
        }

        // Create a LocalSolver array to be able to access it with "at" operator
        LSExpression tasksArray = model.array();
        for (int i = 0; i < nbTasks; ++i) {
            tasksArray.addOperand(tasks[i]);
        }

        // Precedence constraints between the tasks
        for (int i = 0; i < nbTasks; ++i) {
            for (int s = 0; s < nbSuccessors[i]; ++s) {
                model.constraint(model.lt(tasks[i], tasks[successors[i][s]]));
            }
        }

        // Makespan: end of the last task
        makespan = model.max();
        for (int i = 0; i < nbTasks; ++i) {
            makespan.addOperand(model.end(tasks[i]));
        }

        // Cumulative resource constraints
        for (int r = 0; r < nbResources; ++r) {
            final int rL = r;
            LSExpression capacityRespected = model.lambdaFunction(t -> {
                LSExpression taskWeight = model.lambdaFunction(i -> {
                    LSExpression rExpr = model.createConstant(rL);
                    return model.prod(
                           model.at(weight, rExpr, i),
                           model.contains(model.at(tasksArray, i), t));
                });
                LSExpression totalWeight = model.sum(resourcesTasks[rL], taskWeight);
                return model.leq(totalWeight, capacity[rL]);
            });
            model.constraint(model.and(model.range(0, makespan), capacityRespected));
        }

        // Minimize the makespan
        model.minimize(makespan);
        model.close();

        // Parameterize the solver
        localsolver.getParam().setTimeLimit(timeLimit);

        localsolver.solve();
    }

    /*
     * Write the solution in a file with the following format:
     * - total makespan
     * - for each task, the task id, the selected resource, the start and end times
     */
    public void writeSolution(String fileName) throws IOException {
        try (PrintWriter output = new PrintWriter(fileName)) {
            System.out.println("Solution written in file " + fileName);

            output.println(makespan.getValue());

            for (int i = 0; i < nbTasks; ++i) {
                output.println(i + " " + taskResource[i].getValue() + " " + tasks[i].getIntervalValue().getStart() + " "
                        + tasks[i].getIntervalValue().getEnd());
            }
        }
    }

    public static void main(String[] args) {
        if (args.length < 1) {
            System.out.println("Usage: java FlexibleCumulative instanceFile [outputFile] [timeLimit]");
            System.exit(1);
        }

        String instanceFile = args[0];
        String outputFile = args.length > 1 ? args[1] : null;
        String strTimeLimit = args.length > 2 ? args[2] : "60";

        try (LocalSolver localsolver = new LocalSolver()) {
            FlexibleCumulative model = new FlexibleCumulative(localsolver, instanceFile);
            model.readInstance(instanceFile);
            model.solve(Integer.parseInt(strTimeLimit));
            if (outputFile != null) {
                model.writeSolution(outputFile);
            }
        } catch (Exception ex) {
            System.err.println(ex);
            ex.printStackTrace();
            System.exit(1);
        }
    }
}
