use io;

/* Read instance data.*/
function input() {
    local usage = "Usage: localsolver flexible_cumulative.lsp inFileName=instanceFile "
            + "[outFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    inFile = io.openRead(inFileName);
    // Number of tasks
    nbTasks = inFile.readInt();
    // Number of resources
    nbResources = inFile.readInt();

    // Maximum capacity of each resource
    capacity[r in 0...nbResources] = inFile.readInt();
    for [i in 0...nbTasks][r in 0...nbResources] {
        // Duration of task i if task i is done by resource r
        taskProcessingTime[i][r] = inFile.readInt();
        // Resource weight of resource r required for task i
        weight[r][i] = inFile.readInt(); 
    }

    for [i in 0...nbTasks] {
        // The number of successors of task i
        nbSuccessors[i] = inFile.readInt();
        // Successors of each task i
        for [s in 0...nbSuccessors[i]] {
            successors[i][s] = inFile.readInt();
        } 
    } 

    // Trivial upper bound for the start times of the tasks
    horizon = sum[i in 0...nbTasks](max[r in 0...nbResources](taskProcessingTime[i][r]));
    
    inFile.close();
}


function model() {
    // Set of tasks done by each resource
    resourcesTasks[r in 0...nbResources] <- set(nbTasks);

    // Only compatible resources can be selected for a task
    for [i in 0...nbTasks][r in 0...nbResources] {
        if (taskProcessingTime[i][r] == 0 && weight[r][i] == 0)
            constraint contains(resourcesTasks[r], i) == 0;
    }

    // All tasks are scheduled on the resources
    constraint partition[r in 0...nbResources](resourcesTasks[r]);

    // For each task, the selected resource
    taskResource[i in 0...nbTasks] <- find(resourcesTasks, i);

    // Interval decisions: time range of each task
    tasks[i in 0...nbTasks] <- interval(0, horizon);

    // Task duration constraints
    for [i in 0...nbTasks] {
        constraint length(tasks[i]) == taskProcessingTime[i][taskResource[i]];
    }

    // Precedence constraints between the tasks
    for [i in 0...nbTasks][s in 0...nbSuccessors[i]] {
        constraint tasks[i] < tasks[successors[i][s]];
    }

    // Makespan: end of the last task
    makespan <- max[i in 0...nbTasks](end(tasks[i]));

    // Cumulative resource constraints
    for [r in 0...nbResources] {
        constraint and(0...makespan, t => 
                sum(resourcesTasks[r], i => weight[r][i] * contains(tasks[i], t)) <= capacity[r]);
    }

    // Minimize the makespan
    minimize makespan;
}

/* Parameterize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 60;
}

/* Write the solution in a file with the following format:
 *  - total makespan
 *  - for each task, the task id, the selected resource, the start and end times */
function output() {
    if (outFileName != nil) {
        outFile = io.openWrite(outFileName);
        println("Solution written in file ", outFileName);
        outFile.println(makespan.value);
        for [i in 0...nbTasks] {
            outFile.println(i, " ", taskResource[i].value, " ", tasks[i].value.start, " ", tasks[i].value.end);
        }
    }
}
